"""
Python RMSD tool
"""

from typing import Any

import numpy as np

from spyrmsd import molecule, rmsd


def coords_from_molecule(mol: molecule.Molecule, center: bool = False) -> np.ndarray:
    """
    Atomic coordinates from molecule.

    Parameters
    ----------
    mol: molecule.Molecule
        Molecule
    center: bool
        Center flag

    Returns
    -------
    np.ndarray
        Atomic coordinates (possibly centred)

    Notes
    -----
    Atomic coordinates are centred according to the center of geometry, not the center
    of mass.
    """

    if center:
        coords = mol.coordinates - mol.center_of_geometry()
    else:
        coords = mol.coordinates

    return coords


def rmsdwrapper(
    molref,
    mols,
    symmetry: bool = True,
    center: bool = False,
    minimize: bool = False,
    strip: bool = True,
    cache: bool = True,
) -> Any:
    """
    Compute RMSD between two molecule.

    Parameters
    ----------
    mol1: molecule.Molecule
        Molecule 1
    mol2: molecule.Molecule
        Molecule 2
    symmetry: bool, optional
        Symmetry-corrected RMSD (using graph isomorphism)
    center: bool, optional
        Center molecules at origin
    minimize: bool, optional
        Minimised RMSD (using the quaternion polynomial method)
    strip: bool, optional
        Strip hydrogen atoms

    Returns
    -------
    List[float]
        RMSDs
    """

    if strip:
        molref.strip()

        for mol in mols:
            mol.strip()

    if minimize:
        center = True

    cref = coords_from_molecule(molref, center)
    cmols = [coords_from_molecule(mol, center) for mol in mols]

    RMSDlist = []

    if symmetry:
        RMSDlist = rmsd.symmrmsd(
            cref,
            cmols,
            molref.atomicnums,
            mols[0].atomicnums,
            molref.adjacency_matrix,
            mols[0].adjacency_matrix,
            center=center,
            minimize=minimize,
            cache=cache,
        )
    else:  # No symmetry
        for c in cmols:
            RMSDlist.append(
                rmsd.rmsd(
                    cref,
                    c,
                    molref.atomicnums,
                    mols[0].atomicnums,
                    center=center,
                    minimize=minimize,
                )
            )

    return RMSDlist


if __name__ == "__main__":

    from spyrmsd import io

    import sys

    import argparse as ap

    parser = ap.ArgumentParser(description="Python RMSD tool.")

    parser.add_argument("reference", type=str, help="Reference file")
    parser.add_argument("molecules", type=str, nargs="+", help="Input file(s)")
    parser.add_argument("-m", "--minimize", action="store_true", help="Minimize (fit)")
    parser.add_argument(
        "-c", "--center", action="store_true", help="Center molecules at origin"
    )
    parser.add_argument("--hydrogens", action="store_true", help="Keep hydrogen atoms")
    parser.add_argument(
        "-n", "--nosymm", action="store_false", help="No graph isomorphism"
    )

    args = parser.parse_args()

    try:
        ref = io.loadmol(args.reference)
    except OSError:
        print("ERROR: Reference file not found.", file=sys.stderr)
        exit(-1)

    # Load all molecules
    try:
        mols = [mol for molfile in args.molecules for mol in io.loadallmols(molfile)]
    except OSError:
        print("ERROR: Molecule file(s) not found.", file=sys.stderr)
        exit(-1)

    # Loop over molecules within fil
    RMSDlist = rmsdwrapper(
        ref,
        mols,
        symmetry=args.nosymm,  # args.nosymm store False
        center=args.center,
        minimize=args.minimize,
        strip=not args.hydrogens,
    )

    for RMSD in RMSDlist:
        print(f"{RMSD:.5f}")



import copy

import numpy as np
import pytest

from spyrmsd import hungarian, molecule
from tests import molecules


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_cost_mtx(mol: molecule.Molecule):

    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    M = hungarian.cost_mtx(mol1.coordinates, mol2.coordinates)

    n = len(mol1)

    assert len(mol2) == n
    assert M.shape == (n, n)
    assert np.all(M >= 0)

    for i in range(n):
        for j in range(n):
            a, b = mol1.coordinates[i, :], mol2.coordinates[j, :]
            ab = b - a
            assert M[i, j] == pytest.approx(np.dot(ab, ab))


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_optimal_assignement_same_molecule(mol: molecule.Molecule):

    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    assert len(mol1) == len(mol2)

    n = len(mol1)

    cost, row_idx, col_idx = hungarian.optimal_assignment(
        mol1.coordinates, mol2.coordinates
    )

    # When molecules are the same, the assignment is perfect and the cost is zero
    assert cost == pytest.approx(0)

    assert len(row_idx) == len(col_idx) == n

    # The perfect assignment corresponds to the original ordering
    assert np.all(row_idx == np.arange(0, n))
    assert np.all(col_idx == np.arange(0, n))



import copy
import os
from collections import defaultdict
from typing import DefaultDict, List, Tuple

import numpy as np
import pytest
import qcelemental as qcel

from spyrmsd import graph, io, molecule, utils
from tests import molecules


# atoms is a list of atomic numbers and atom counts
@pytest.mark.parametrize(
    "mol, atoms",
    [
        (molecules.benzene, [(1, 6), (6, 6)]),
        (molecules.ethanol, [(1, 6), (6, 2), (8, 1)]),
        (molecules.dialanine, [(1, 12), (6, 6), (7, 2), (8, 3)]),
    ],
)
def test_load(mol: molecule.Molecule, atoms: List[Tuple[int, int]]) -> None:

    n = sum([n_atoms for _, n_atoms in atoms])

    assert len(mol) == n
    assert mol.atomicnums.shape == (n,)
    assert mol.coordinates.shape == (n, 3)

    # Count number of atoms of different elements
    atomcount: DefaultDict[int, int] = defaultdict(int)
    for atomicnum in mol.atomicnums:
        atomcount[atomicnum] += 1

    assert len(atomcount) == len(atoms)

    for Z, n_atoms in atoms:
        assert atomcount[Z] == n_atoms


def test_loadall() -> None:

    path = os.path.join(molecules.molpath, "1cbr_docking.sdf")

    mols = io.loadall(path)

    assert len(mols) == 10


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_molecule_translate(mol: molecule.Molecule) -> None:

    mt = copy.deepcopy(mol)

    t = np.array([0.5, 1.1, -0.1])
    mt.translate(t)

    for tcoord, coord in zip(mt.coordinates, mol.coordinates):
        assert np.allclose(tcoord - t, coord)


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_molecule_rotate_z(mol: molecule.Molecule) -> None:

    z_axis = np.array([0, 0, 1])

    for angle in [0, 45, 90]:

        rotated = np.zeros((len(mol), 3))
        for i, coord in enumerate(mol.coordinates):
            rotated[i] = utils.rotate(coord, angle, z_axis, units="deg")

        mol.rotate(angle, z_axis, units="deg")

        assert np.allclose(mol.coordinates, rotated)

        # Reset
        mol.rotate(-angle, z_axis, units="deg")


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_molecule_rotate(mol: molecule.Molecule) -> None:

    axis = np.random.rand(3)

    for angle in np.random.rand(10) * 180:

        rotated = np.zeros((len(mol), 3))
        for i, coord in enumerate(mol.coordinates):
            rotated[i] = utils.rotate(coord, angle, axis, units="deg")

        mol.rotate(angle, axis, units="deg")

        assert np.allclose(mol.coordinates, rotated)

        # Reset
        mol.rotate(-angle, axis, units="deg")


def test_molecule_center_of_geometry_benzene() -> None:

    mol = molecules.benzene

    assert np.allclose(mol.center_of_geometry(), np.zeros(3))


def test_molecule_center_of_mass_benzene() -> None:

    mol = molecules.benzene

    assert np.allclose(mol.center_of_mass(), np.zeros(3))


def test_molecule_center_of_mass_H2() -> None:

    atomicnums = [1, 1]
    coordinates = [[0, 0, -1], [0, 0, 1]]

    mol = molecule.Molecule(atomicnums, coordinates)

    assert np.allclose(mol.center_of_mass(), np.zeros(3))


def test_molecule_center_of_mass_HF() -> None:

    atomicnums = [1, 9]
    coordinates = [[0, 0, -1], [0, 0, 1]]

    H_mass = qcel.periodictable.to_mass(1)
    F_mass = qcel.periodictable.to_mass(9)

    z_com = (-H_mass + F_mass) / (H_mass + F_mass)

    mol = molecule.Molecule(atomicnums, coordinates)

    assert np.allclose(mol.center_of_mass(), np.array([0, 0, z_com]))


@pytest.mark.parametrize(
    "mol, n_atoms, stripped",
    [
        (molecules.benzene, 12, 6),
        (molecules.ethanol, 9, 6),
        (molecules.dialanine, 23, 12),
    ],
)
def test_molecule_strip(mol: molecule.Molecule, n_atoms: int, stripped: int) -> None:

    m = copy.deepcopy(mol)

    assert len(m) == n_atoms

    m.strip()

    assert len(m) == n_atoms - stripped


@pytest.mark.parametrize(
    "mol, n_bonds",
    [(molecules.benzene, 12), (molecules.ethanol, 8), (molecules.dialanine, 22)],
)
def test_graph_from_adjacency_matrix(mol: molecule.Molecule, n_bonds: int) -> None:

    G = mol.to_graph()

    assert graph.num_vertices(G) == len(mol)
    assert graph.num_edges(G) == n_bonds

    for idx, atomicnum in enumerate(mol.atomicnums):
        assert graph.vertex_property(G, "atomicnum", idx) == atomicnum


@pytest.mark.parametrize(
    "mol, n_bonds",
    [(molecules.benzene, 12), (molecules.ethanol, 8), (molecules.dialanine, 22)],
)
def test_graph_from_atomic_coordinates_perception(
    mol: molecule.Molecule, n_bonds: int
) -> None:

    m = copy.deepcopy(mol)

    m.adjacency_matrix = None
    m.G = None

    with pytest.warns(UserWarning):

        # Uses automatic bond perception
        G = m.to_graph()

        assert graph.num_vertices(G) == len(m)
        assert graph.num_edges(G) == n_bonds

        for idx, atomicnum in enumerate(mol.atomicnums):
            assert graph.vertex_property(G, "atomicnum", idx) == atomicnum



import warnings
from typing import Any, List, Optional, Tuple, Union

import graph_tool as gt
import numpy as np
from graph_tool import generation, topology


def graph_from_adjacency_matrix(
    adjacency_matrix: Union[np.ndarray, List[List[int]]],
    atomicnums: Optional[Union[np.ndarray, List[int]]] = None,
):
    """
    Graph from adjacency matrix.

    Parameters
    ----------
    adjacency_matrix: Union[np.ndarray, List[List[int]]]
        Adjacency matrix
    atomicnums: Union[np.ndarray, List[int]], optional
        Atomic numbers

    Returns
    -------
    Graph
        Molecular graph

    Notes
    -----
    It the atomic numbers are passed, they are used as node attributes.
    """

    # Get upper triangular adjacency matrix
    adj = np.triu(adjacency_matrix)

    G = gt.Graph(directed=False)
    G.add_edge_list(np.transpose(adj.nonzero()))

    if atomicnums is not None:
        vprop = G.new_vertex_property("short")  # Create property map (of C type short)
        vprop.a = atomicnums  # Assign atomic numbers to property map array
        G.vertex_properties["atomicnum"] = vprop  # Set property map

    return G


def match_graphs(G1, G2) -> List[Tuple[List[int], List[int]]]:
    """
    Compute graph isomorphisms.

    Parameters
    ----------
    G1:
        Graph 1
    G2:
        Graph 2

    Returns
    -------
    List[Tuple[List[int],List[int]]]
        All possible mappings between nodes of graph 1 and graph 2 (isomorphisms)

    Raises
    ------
    ValueError
        If the graphs `G1` and `G2` are not isomorphic
    """

    try:
        maps = topology.subgraph_isomorphism(
            G1,
            G2,
            vertex_label=(
                G1.vertex_properties["atomicnum"],
                G2.vertex_properties["atomicnum"],
            ),
            subgraph=False,
        )
    except KeyError:  # No "atomicnum" vertex property
        warnings.warn(
            "No atomic number information stored on nodes. "
            + "Node matching is not performed..."
        )

        maps = topology.subgraph_isomorphism(G1, G2, subgraph=False)

    # Check if graphs are actually isomorphic
    if len(maps) == 0:
        # TODO: Create a new exception
        raise ValueError(f"Graphs {G1} and {G2} are not isomorphic.")

    n = num_vertices(G1)

    # Extract all isomorphisms in a list
    return [(np.arange(0, n, dtype=int), m.a) for m in maps]


def vertex_property(G, vproperty: str, idx: int) -> Any:
    """
    Get vertex (node) property from graph

    Parameters
    ----------
    G:
        Graph
    vproperty: str
        Vertex property name
    idx: int
        Vertex index

    Returns
    -------
    Any
        Vertex property value
    """
    return G.vertex_properties[vproperty][idx]


def num_vertices(G) -> int:
    """
    Number of vertices

    Parameters
    ----------
    G:
        Graph

    Returns
    -------
    int
        Number of vertices (nodes)
    """
    return G.num_vertices()


def num_edges(G) -> int:
    """
    Number of edges

    Parameters
    ----------
    G:
        Graph

    Returns
    -------
    int
        Number of edges
    """
    return G.num_edges()


def lattice(n1: int, n2: int):
    """
    Build 2D lattice graph

    Parameters
    ----------
    n1: int
        Number of nodes in dimension 1
    n2: int
        Number of nodes in dimension 2

    Returns
    -------
    Graph
        Lattice graph
    """
    return generation.lattice((n1, n2))


def cycle(n):
    """
    Build cycle graph

    Parameters
    ----------
    n: int
        Number of nodes

    Returns
    -------
    Graph
        Cycle graph
    """
    return generation.circular_graph(n)



from typing import List, Tuple

import numpy as np

from spyrmsd import molecule, utils

try:
    # 3.0
    from openbabel import openbabel as ob
    from openbabel import pybel
except ImportError:
    # 2.0
    import openbabel as ob
    import pybel


def load(fname: str):
    """
    Load molecule from file.

    Parameters
    ----------
    fname: str
        File name

    Returns
    -------
    Molecule
    """

    fmt = utils.molformat(fname)

    obmol = next(pybel.readfile(fmt, fname))

    return obmol


def loadall(fname: str):
    """
    Load molecules from file.

    Parameters
    ----------
    fname: str
        File name

    Returns
    -------
    List of molecules
    """

    fmt = utils.molformat(fname)

    obmols = [obmol for obmol in pybel.readfile(fmt, fname)]

    # FIXME: Special handling for multi-model PDB files
    # See OpenBabel Issue #2097
    if fmt == "pdb":
        if len(obmols) > 1:  # Multi-model PDB file
            obmols = obmols[:-1]

    return obmols


def adjacency_matrix(mol) -> np.ndarray:
    """
    Adjacency matrix from OpenBabel molecule.

    Parameters
    ----------
    mol:
        Molecule

    Returns
    -------
    np.ndarray
        Adjacency matrix of the molecule
    """

    n = len(mol.atoms)

    # Pre-allocate memory for  the adjacency matrix
    A = np.zeros((n, n), dtype=int)

    # Loop over molecular bonds
    for bond in ob.OBMolBondIter(mol.OBMol):
        # Bonds are 1-indexed
        i: int = bond.GetBeginAtomIdx() - 1
        j: int = bond.GetEndAtomIdx() - 1

        # A molecular graph is undirected
        A[i, j] = A[j, i] = 1

    return A


def to_molecule(mol, adjacency: bool = True) -> molecule.Molecule:
    """
    Transform molecule to `pyrmsd` molecule.

    Parameters
    ----------
    mol:
        Molecule
    adjacency: boolean, optional
        Flag to decide wether to build the adjacency matrix from molecule

    Returns
    -------
    pyrmsd.molecule.Molecule
        `pyrmsd` molecule
    """

    n = len(mol.atoms)

    atomicnums = np.zeros(n, dtype=int)
    coordinates = np.zeros((n, 3))

    for i, atom in enumerate(mol.atoms):
        atomicnums[i] = atom.atomicnum
        coordinates[i] = atom.coords

    if adjacency:
        A = adjacency_matrix(mol)

    return molecule.Molecule(atomicnums, coordinates, A)


def numatoms(mol) -> int:
    """
    Number of atoms.

    Parameters
    ----------
    mol:
        Molecule

    Returns
    -------
    int
        Number of atoms
    """
    return mol.OBMol.NumAtoms()


def numbonds(mol) -> int:
    """
    Number of bonds.

    Parameters
    ----------
    mol:
        Molecule

    Returns
    -------
    int
        Number of bonds
    """
    return mol.OBMol.NumBonds()


def bonds(mol) -> List[Tuple[int, int]]:
    """
    List of bonds

    Parameters
    ----------
    mol:
        Molecule

    Returns
    -------
    List[Tuple[int, int]]
        List of bonds

    Notes
    -----
    A bond is defined by a tuple of (0-based) indices of two atoms.
    """
    b = []

    for bond in ob.OBMolBondIter(mol.OBMol):
        i = bond.GetBeginAtomIdx() - 1
        j = bond.GetEndAtomIdx() - 1

        b.append((i, j))

    return b



from typing import Tuple

import numpy as np
from scipy import optimize

from .due import due, Doi

due.cite(
    Doi("10.1107/S0108767305015266"),  # lgtm[py/procedure-return-value-used]
    path="spyrmsd.qcp",
    description="QCP method",
)


def M_mtx(A: np.ndarray, B: np.ndarray) -> np.ndarray:
    """
    Compute inner product between coordinate matrices.

    Parameters
    ----------
    A: numpy.ndarray
        Coordinates `A`
    B: numpy.ndarray
        Coordinates `B`

    Returns
    -------
    numpy.ndarray
        Inner product of the coordinate matrices `A` and `B`

    Notes
    -----
    The inner product of the coordinate matrices `A` and `B` corresponds to the matrix
    :math:`\\mathbf{M}`. [1]_

    If :math:`S_{xy}` is defined as

    .. math:: S_{xy} = \\sum_i^N x_{B,i} y_{A,i}

    then :math:`\\mathbf{M}` is the :math:`3\\times 3` matrix given by

    .. math::
       \\begin{pmatrix}
            S_{xx} & S_{xy} & S_{xz} \\\\
            S_{yx} & S_{yy} & S_{yz} \\\\
            S_{zx} & S_{zy} & S_{zz} \\\\
       \\end{pmatrix}

    .. [1] D. L. Theobald, *Rapid calculation of RMSDs using a quaternion-based
       characteristic polynomial*, Acta Crys. A **61**, 478-480 (2005).
    """
    return B.T @ A


def K_mtx(M):
    """
    Compute symmetric key matrix.

    Parameters
    ----------
    M : numpy.ndarray
        Inner product between coordinate matrices

    Returns
    -------
    numpy.ndarray
        Symmetric key matrix

    Notes
    -----
    The symmetric key matrix corresponds to the matrix :math:`\\mathbf{K}`. [2]_

    If :math:`S_{xy}` is defined as

    .. math:: S_{xy} = \\sum_i^N x_{B,i} y_{A,i}

    then :math:`\\mathbf{K}` is the :math:`4\\times 4` symmetric matrix given by

    .. math::
       \\begin{pmatrix}
            S_{xx} + S_{yy} + S_{zz} & S_{yz} - S_{zy} & S_{zx} - S_{xz} & S_{xy} - S_{yx} \\\\
            & S_{xx} - S_{yy} - S_{zz} & S_{xy} + S_{yx} & S_{zx} + S_{xz}\\\\
            & & -S_{xx} + S_{yy} - S_{zz} & S_{yz} - S_{zy} \\\\
            &  & & -S_{xx} - S_{yy} + S_{zz} \\\\
       \\end{pmatrix}

    .. [2] D. L. Theobald, *Rapid calculation of RMSDs using a quaternion-based
       characteristic polynomial*, Acta Crys. A **61**, 478-480 (2005).
    """

    assert M.shape == (3, 3)

    S_xx = M[0, 0]
    S_xy = M[0, 1]
    S_xz = M[0, 2]
    S_yx = M[1, 0]
    S_yy = M[1, 1]
    S_yz = M[1, 2]
    S_zx = M[2, 0]
    S_zy = M[2, 1]
    S_zz = M[2, 2]

    # p = plus, m = minus
    S_xx_yy_zz_ppp = S_xx + S_yy + S_zz
    S_yz_zy_pm = S_yz - S_zy
    S_zx_xz_pm = S_zx - S_xz
    S_xy_yx_pm = S_xy - S_yx
    S_xx_yy_zz_pmm = S_xx - S_yy - S_zz
    S_xy_yx_pp = S_xy + S_yx
    S_zx_xz_pp = S_zx + S_xz
    S_xx_yy_zz_mpm = -S_xx + S_yy - S_zz
    S_yz_zy_pp = S_yz + S_zy
    S_xx_yy_zz_mmp = -S_xx - S_yy + S_zz

    return np.array(
        [
            [S_xx_yy_zz_ppp, S_yz_zy_pm, S_zx_xz_pm, S_xy_yx_pm],
            [S_yz_zy_pm, S_xx_yy_zz_pmm, S_xy_yx_pp, S_zx_xz_pp],
            [S_zx_xz_pm, S_xy_yx_pp, S_xx_yy_zz_mpm, S_yz_zy_pp],
            [S_xy_yx_pm, S_zx_xz_pp, S_yz_zy_pp, S_xx_yy_zz_mmp],
        ]
    )


def coefficients(M: np.ndarray, K: np.ndarray) -> Tuple[float, float, float]:
    """
    Compute quaternion polynomial coefficients.

    Parameters
    ----------
    M : numpy.ndarray
        Inner product between coordinate matrices
    K: numpy.ndarray
        Symmetric key matrix

    Returns
    -------
    Tuple[float, float, float]
        Quaternion polynomial coefficients

    Notes
    _____
    Returns only :math:`\\mathbf{M}`- and :math:`\\mathbf{K}`-dependent coefficients
    are returned. :math:`c_4=1` and :math:`c_3=0` are not returned.

    The :math:`\\mathbf{M}`- and :math:`\\mathbf{K}`-dependent quaternion polynomial
    coefficients are given by

    .. math:: c_2 = -2 \\text{ tr}\\left(\\mathbf{M}^T\\mathbf{M}\\right)

    .. math:: c_1 = -8 \\text{ det}(\\mathbf{M})

    .. math:: c_0 = \\text{ det}(\\mathbf{K})

    """

    c2 = -2 * np.trace(M.T @ M)
    c1 = -8 * np.linalg.det(M)  # TODO: Slow?
    c0 = np.linalg.det(K)  # TODO: Slow?

    return c2, c1, c0


def lambda_max(Ga: float, Gb: float, c2: float, c1: float, c0: float) -> float:
    """
    Find largest root of the quaternion polynomial.

    Parameters
    ----------
    Ga: float
        Inner product of structure A
    Gb:
        Inner product of structure B
    c2:
        Coefficient :math:`c_2` of the quaternion polynomial
    c1:
        Coefficient :math:`c_1` of the quaternion polynomial
    c0:
        Coefficient :math:`c_0` of the quaternion polynomial

    Returns
    -------
    float
        Largest root of the quaternion polynomial (:math:`\\lambda_\\text{max}`)
    """

    def P(x):
        """
        Quaternion polynomial
        """
        return x ** 4 + c2 * x ** 2 + c1 * x + c0

    def dP(x):
        """
        Fist derivative of the quaternion polynomial
        """
        return 4 * x ** 3 + 2 * c2 * x + c1

    x0 = (Ga + Gb) / 2.0

    return optimize.newton(P, x0, fprime=dP)


def qcp_rmsd(A: np.ndarray, B: np.ndarray) -> float:
    """
    Compute RMSD using the quaternion polynomial method.

    Parameters
    ----------
    A: numpy.ndarray
        Coordinates of structure A
    B: numpy.ndarray
        Coordinates of structure B

    Returns
    -------
    float
        RMSD between structures `A` and `B`

    Raises
    ------
    AssertionError
        If the shape of structures `A` and `B` is different

    Notes
    -----
    If the structures `A` and `B` can be superimposed exactly (i.e. they differ only
    by center-of-mass translations and rotations), we have

        .. math:: G_a + G_b = 2 \\lambda_\\text{max}

    This means that :math:`s = G_a + G_bb - 2 * \\lambda_\\text{max}` can become
    negative because of numerical errors and therefore :math:`\\sqrt{s}` fails.
    In order to avoid this problem, the final RMSD is set to :math:`0`
    if :math:`|s| < 10^{-12}`.
    """

    assert A.shape == B.shape

    N = A.shape[0]

    Ga = np.trace(A.T @ A)
    Gb = np.trace(B.T @ B)

    M = M_mtx(A, B)
    K = K_mtx(M)

    c2, c1, c0 = coefficients(M, K)

    l_max = lambda_max(Ga, Gb, c2, c1, c0)

    s = Ga + Gb - 2 * l_max
    if abs(s) < 1e-12:  # Avoid numerical errors when Ga + Gb = 2 * l_max
        rmsd = 0
    else:
        rmsd = np.sqrt(s / N)

    return rmsd



import warnings
from typing import List, Optional, Union

import numpy as np
import qcelemental as qcel

from spyrmsd import graph, utils


class Molecule:
    def __init__(
        self,
        atomicnums: Union[np.ndarray, List[int]],
        coordinates: Union[np.ndarray, List[List[float]]],
        adjacency_matrix: Union[np.ndarray, List[List[int]]] = None,
    ) -> None:
        """
        Molecule initialisation.

        Parameters
        ----------
        atomicnums: Union[np.ndarray, List[int]]
            Atomic numbers
        coordinates: Union[np.ndarray, List[List[float]]]
            Atomic coordinates
        adjacency_matrix: Union[np.ndarray, List[List[int]]], optional
            Molecular graph adjacency matrix

        Notes
        -----

        A molecule is built from atomic numbers and atomic coordinates only.
        Optionally, a good representation of the molecular graph (obtained with
        OpenBabel or RDKit) can be stored as adjacency matrix.
        """

        atomicnums = np.asarray(atomicnums, dtype=int)
        coordinates = np.asarray(coordinates, dtype=float)

        self.natoms: int = len(atomicnums)

        assert atomicnums.shape == (self.natoms,)
        assert coordinates.shape == (self.natoms, 3)

        self.atomicnums = atomicnums
        self.coordinates = coordinates

        self.stripped: bool = np.all(atomicnums != 1)

        if adjacency_matrix is not None:
            self.adjacency_matrix = np.asarray(adjacency_matrix, dtype=int)

        # Molecular graph
        self.G = None

        self.masses: Optional[List[float]] = None

    def translate(self, vector: np.ndarray) -> None:
        """
        Translate molecule.

        Parameters
        ----------
        vector: np.ndarray
            Translation vector (in 3D)
        """
        assert len(vector) == 3
        self.coordinates += vector

    def rotate(self, angle: float, axis: np.ndarray, units: str = "rad") -> None:
        """
        Rotate molecule.

        Parameters
        ----------
        angle: float
            Rotation angle
        axis: np.ndarray
            Axis of rotation (in 3D)
        units: {"rad", "deg"}
            Units of the angle (radians `rad` or degrees `deg`)
        """
        self.coordinates = utils.rotate(self.coordinates, angle, axis, units)

    def center_of_mass(self) -> np.ndarray:
        """
        Center of mass.

        Returns
        -------
        np.ndarray
            Center of mass

        Notes
        -----
        Atomic masses are cached.
        """

        # Get masses and cache them
        if self.masses is None:
            self.masses = [qcel.periodictable.to_mass(anum) for anum in self.atomicnums]

        return np.average(self.coordinates, axis=0, weights=self.masses)

    def center_of_geometry(self) -> np.ndarray:
        """
        Center of geometry.

        Returns
        -------
        np.ndarray
            Center of geometry
        """
        return utils.center_of_geometry(self.coordinates)

    # TODO: Change name (to stripH)
    def strip(self) -> None:
        """
        Strip hydrogen atoms.
        """

        if not self.stripped:
            idx = self.atomicnums != 1  # Non-hydrogen atoms

            # Strip
            self.atomicnums = self.atomicnums[idx]
            self.coordinates = self.coordinates[idx, :]

            # Update number of atoms
            self.natoms = len(self.atomicnums)

            # Update adjacency matrix
            if self.adjacency_matrix is not None:
                self.adjacency_matrix = self.adjacency_matrix[np.ix_(idx, idx)]

            # Reset molecular graph when stripping
            self.G = None

            self.stripped = True

    def to_graph(self):
        """
        Convert molecule to graph.

        Returns
        -------
        Graph
            Molecular graph.

        Notes
        -----
        If the molecule does not have an associated adjacency matrix, a simple
        bond perception is used.

        The molecular graph is cached.
        """
        if self.G is None:
            if self.adjacency_matrix is not None:
                self.G = graph.graph_from_adjacency_matrix(
                    self.adjacency_matrix, self.atomicnums
                )
            else:
                warnings.warn(
                    "Molecule was not initialized with an adjacency matrix. "
                    + "Using bond perception..."
                )

                # Automatic bond perception (with very simple rule)
                self.adjacency_matrix = graph.adjacency_matrix_from_atomic_coordinates(
                    self.atomicnums, self.coordinates
                )

                self.G = graph.graph_from_adjacency_matrix(
                    self.adjacency_matrix, self.atomicnums
                )

        return self.G

    def __len__(self) -> int:
        """
        Molecule size.

        Returns
        -------
        int
            Number of atoms within the molecule
        """
        return self.natoms



import random

import numpy as np
import pytest

from spyrmsd import utils


@pytest.mark.parametrize("ext", ["pdf", "smi", "xyz"])
def test_format(ext: str) -> None:
    for fname in ["test", "root/test", "root/test.test"]:
        fmt = utils.format(f"{fname}.{ext}")
        assert fmt == ext


@pytest.mark.parametrize(
    "extin, extout", [("smi", "smi"), ("pdb", "pdb"), ("xyz", "XYZ")]
)
def test_molformat(extin: str, extout: str) -> None:

    for fname in ["test", "root/test", "root/test.test"]:
        fmt = utils.molformat(f"{fname}.{extin}")
        assert fmt == extout


@pytest.mark.parametrize(
    "deg, rad",
    [(0, 0), (90, np.pi / 2), (180, np.pi), (270, 3 * np.pi / 2), (360, 2 * np.pi)],
)
def test_deg_to_rad(deg: float, rad: float) -> None:

    assert utils.deg_to_rad(deg) == pytest.approx(rad)


def test_rotate_invalid():

    with pytest.raises(ValueError):
        utils.rotate(np.array([1, 0, 0]), 0, np.array([0, 0, 1]), units="none")


@pytest.mark.parametrize("z", [np.array([0, 0, random.random()]) for _ in range(10)])
def test_rotate_z(z: np.ndarray) -> None:

    for deg, rad in [(0, 0), (45, np.pi / 4), (90, np.pi / 2)]:
        v_deg = utils.rotate(np.array([1, 0, 0]), deg, z, units="deg")
        v_rad = utils.rotate(np.array([1, 0, 0]), rad, z, units="rad")

        assert np.allclose(v_deg, v_rad)
        assert np.allclose(v_deg, np.array([np.cos(rad), np.sin(rad), 0]))



import numpy as np
import pytest

from spyrmsd import graph, io, molecule
from tests import molecules


@pytest.mark.parametrize(
    "mol, n_bonds",
    [(molecules.benzene, 12), (molecules.ethanol, 8), (molecules.dialanine, 22)],
)
def test_adjacency_matrix_from_atomic_coordinates(
    mol: molecule.Molecule, n_bonds: int
) -> None:

    A = graph.adjacency_matrix_from_atomic_coordinates(mol.atomicnums, mol.coordinates)

    G = graph.graph_from_adjacency_matrix(A)

    assert graph.num_vertices(G) == len(mol)
    assert graph.num_edges(G) == n_bonds


@pytest.mark.parametrize("mol", molecules.allobmolecules)
def test_adjacency_matrix_from_mol(mol) -> None:

    natoms = io.numatoms(mol)
    nbonds = io.numbonds(mol)

    A = io.adjacency_matrix(mol)

    assert A.shape == (natoms, natoms)
    assert np.alltrue(A == A.T)
    assert np.sum(A) == nbonds * 2

    for i, j in io.bonds(mol):

        assert A[i, j] == 1


@pytest.mark.parametrize("mol", molecules.allobmolecules)
def test_graph_from_adjacency_matrix(mol) -> None:

    natoms = io.numatoms(mol)
    nbonds = io.numbonds(mol)

    A = io.adjacency_matrix(mol)

    assert A.shape == (natoms, natoms)
    assert np.alltrue(A == A.T)
    assert np.sum(A) == nbonds * 2

    G = graph.graph_from_adjacency_matrix(A)

    assert graph.num_vertices(G) == natoms
    assert graph.num_edges(G) == nbonds


@pytest.mark.parametrize(
    "rawmol, mol", zip(molecules.allobmolecules, molecules.allmolecules)
)
def test_graph_from_adjacency_matrix_atomicnums(rawmol, mol) -> None:

    natoms = io.numatoms(rawmol)
    nbonds = io.numbonds(rawmol)

    A = io.adjacency_matrix(rawmol)

    assert len(mol) == natoms
    assert mol.adjacency_matrix.shape == (natoms, natoms)
    assert np.alltrue(mol.adjacency_matrix == A)
    assert np.sum(mol.adjacency_matrix) == nbonds * 2

    G = mol.to_graph()

    assert graph.num_vertices(G) == natoms
    assert graph.num_edges(G) == nbonds

    for idx, atomicnum in enumerate(mol.atomicnums):
        assert graph.vertex_property(G, "atomicnum", idx) == atomicnum


@pytest.mark.parametrize(
    "G1, G2",
    [
        *[(graph.lattice(n, n), graph.lattice(n, n)) for n in range(2, 5)],
        *[(graph.cycle(n), graph.cycle(n)) for n in range(2, 5)],
    ],
)
def test_match_graphs_isomorphic(G1, G2) -> None:

    with pytest.warns(UserWarning):
        isomorphisms = graph.match_graphs(G1, G2)

    assert len(isomorphisms) != 0


@pytest.mark.parametrize(
    "G1, G2",
    [
        *[(graph.lattice(n, n), graph.lattice(n + 1, n)) for n in range(2, 5)],
        *[(graph.cycle(n), graph.cycle(n + 1)) for n in range(1, 5)],
    ],
)
def test_match_graphs_not_isomorphic(G1, G2) -> None:

    with pytest.raises(ValueError), pytest.warns(UserWarning):
        graph.match_graphs(G1, G2)



try:
    from spyrmsd.graphs.gt import (
        graph_from_adjacency_matrix,
        match_graphs,
        vertex_property,
        num_vertices,
        num_edges,
        lattice,
        cycle,
    )

except ImportError:
    try:
        from spyrmsd.graphs.nx import (
            graph_from_adjacency_matrix,
            match_graphs,
            vertex_property,
            num_vertices,
            num_edges,
            lattice,
            cycle,
        )
    except ImportError:
        raise ImportError("graph_tool or NetworkX libraries not found.")

__all__ = [
    "graph_from_adjacency_matrix",
    "match_graphs",
    "vertex_property",
    "num_vertices",
    "num_edges",
    "lattice",
    "cycle",
    "adjacency_matrix_from_atomic_coordinates",
]

import numpy as np
import qcelemental as qcel

# TODO: Move elsewhere?
connectivity_tolerance: float = 0.4


def adjacency_matrix_from_atomic_coordinates(
    atomicnums: np.ndarray, coordinates: np.ndarray
) -> np.ndarray:
    """
    Compute adjacency matrix from atomic coordinates.

    Parameters
    ----------
    atomicnums: numpy.ndarray
        Atomic numbers
    coordinates: numpy.ndarray
        Atomic coordinates

    Returns
    -------
    numpy.ndarray
        Adjacency matrix

    Notes
    -----

    This function is based on an automatic bond perception algorithm: two
    atoms are considered to be bonded when their distance is smaller than
    the sum of their covalent radii plus a tolerance value. [3]_

    .. warning::
        The automatic bond perceptron rule implemented in this functions
        is very simple and only depends on atomic coordinates. Use
        with care!

    .. [3] E. C. Meng and R. A. Lewis, *Determination of molecular topology and atomic
       hybridization states from heavy atom coordinates*, J. Comp. Chem. **12**, 891-898
       (1991).
    """

    n = len(atomicnums)

    assert coordinates.shape == (n, 3)

    A = np.zeros((n, n))

    for i in range(n):
        r_i = qcel.covalentradii.get(atomicnums[i], units="angstrom")

        for j in range(i + 1, n):
            r_j = qcel.covalentradii.get(atomicnums[i], units="angstrom")

            distance = np.sqrt(np.sum((coordinates[i] - coordinates[j]) ** 2))

            if distance < (r_i + r_j + connectivity_tolerance):
                A[i, j] = A[j, i] = 1

    return A



import os

import pytest

from spyrmsd import io

fdir = os.path.dirname(os.path.abspath(__file__))
molpath = os.path.join(fdir, "data/molecules/")


@pytest.mark.parametrize(
    "molfile, natoms, nbonds",
    [("benzene.sdf", 12, 12), ("ethanol.sdf", 9, 8), ("dialanine.sdf", 23, 22)],
)
def test_load_sdf(molfile, natoms: int, nbonds: int) -> None:

    m = io.load(os.path.join(molpath, molfile))

    assert io.numatoms(m) == natoms
    assert io.numbonds(m) == nbonds


@pytest.mark.parametrize(
    "molfile, natoms, nbonds", [("1cbr_ligand.mol2", 49, 49)],
)
def test_load_mol2(molfile, natoms: int, nbonds: int) -> None:

    m = io.load(os.path.join(molpath, molfile))

    assert io.numatoms(m) == natoms
    assert io.numbonds(m) == nbonds


@pytest.mark.parametrize(
    "molfile, natoms, nbonds",
    [("trp0.pdb", 217, 224), ("trp1.pdb", 217, 224), ("trp2.pdb", 217, 224)],
)
def test_load_pdb(molfile, natoms: int, nbonds: int) -> None:

    m = io.load(os.path.join(molpath, molfile))

    assert io.numatoms(m) == natoms
    assert io.numbonds(m) == nbonds


@pytest.mark.parametrize(
    "molfile, natoms, nbonds", [("1cbr_docking.sdf", 22, 22)],
)
def test_loadall_sdf(molfile, natoms: int, nbonds: int) -> None:

    ms = io.loadall(os.path.join(molpath, molfile))

    assert len(ms) == 10

    for m in ms:
        assert io.numatoms(m) == natoms
        assert io.numbonds(m) == nbonds


@pytest.mark.parametrize(
    "molfile, natoms, nbonds", [("1cbr_docking.mol2", 22, 22)],
)
def test_loadall_mol2(molfile, natoms: int, nbonds: int) -> None:

    try:
        ms = io.loadall(os.path.join(molpath, molfile))

        assert len(ms) == 10

        for m in ms:
            assert io.numatoms(m) == natoms
            assert io.numbonds(m) == nbonds
    except NotImplementedError:  # Mol2MolSupplier in RDkit is not supported
        pass  # TODO: Warning


@pytest.mark.parametrize(
    "molfile, natoms, nbonds", [("1cbr_docking.pdb", 22, 22)],
)
def test_loadall_pdb(molfile, natoms: int, nbonds: int) -> None:

    try:
        ms = io.loadall(os.path.join(molpath, molfile))

        assert len(ms) == 10

        for m in ms:
            assert io.numatoms(m) == natoms
            assert io.numbonds(m) == nbonds

    except NotImplementedError:  # PDBMolSupplier in RDkit is not supported
        pass  # TODO: Warning


def test_loadall_pdb_single_model() -> None:
    """
    Test load_all function for PDB files when only a single model is present.
    """

    try:
        ms = io.loadall(os.path.join(molpath, "1a99_ligand.pdb"))

        assert len(ms) == 1

        assert io.numatoms(ms[0]) == 20
        assert io.numbonds(ms[0]) == 19

    except NotImplementedError:  # PDBMolSupplier in RDkit is not supported
        pass  # TODO: Warning


@pytest.mark.parametrize(
    "molfile, natoms", [("benzene.sdf", 12), ("ethanol.sdf", 9), ("dialanine.sdf", 23)],
)
def test_loadmol_sdf(molfile, natoms: int) -> None:

    m = io.loadmol(os.path.join(molpath, molfile))

    assert len(m) == natoms


@pytest.mark.parametrize(
    "molfile, natoms", [("1cbr_ligand.mol2", 49)],
)
def test_loadmol_mol2(molfile, natoms: int) -> None:

    m = io.loadmol(os.path.join(molpath, molfile))

    assert len(m) == natoms


@pytest.mark.parametrize(
    "molfile, natoms", [("1cbr_docking.sdf", 22)],
)
def test_loadallmols_sdf(molfile, natoms: int) -> None:

    ms = io.loadallmols(os.path.join(molpath, molfile))

    assert len(ms) == 10

    for m in ms:
        assert len(m) == natoms



import copy
import sys
from typing import List

import pytest

from spyrmsd import spyrmsd
from tests import molecules


def test_spyrmsd_imported():
    assert "spyrmsd" in sys.modules


# Results obtained with MDAnalysis
#   trp0 = mda.Universe("trp0.pdb")
#   c0 = trp0.coord -trp0.select_atoms("protein").center_of_geometry()
#   for i in [1,2,3,4,5]:
#       trp = mda.Universe(f"trp{i}.pdb")
#       rmsd_dummy = mda.analysis.rms.rmsd(trp.coord, trp0.coord)
#       c = trp.coord - trp.select_atoms("protein").center_of_geometry()
#       _, rmsd_min = align.rotation_matrix(tc, tc0)
#       print(rmsd_dummy, rmsd_min)
@pytest.mark.parametrize(
    "minimize, referenceRMSDs",
    [
        (
            False,  # No minimize: dummy RMSD
            [
                4.812480551076202,
                6.772045449820714,
                9.344911262612964,
                9.772939589989000,
                8.901837608843241,
            ],
        ),
        (
            True,  # Minimize: QCP
            [
                1.6578281551053196,
                1.7175638492348284,
                1.5946081072641485,
                2.1234944939308220,
                2.4894805175766606,
            ],
        ),
    ],
)
def test_rmsdwrapper_nosymm_protein(minimize: bool, referenceRMSDs: List[float]):

    mol0 = copy.deepcopy(molecules.trp[0])
    mols = [copy.deepcopy(mol) for mol in molecules.trp[1:]]

    RMSDs = spyrmsd.rmsdwrapper(
        mol0, mols, symmetry=False, minimize=minimize, strip=False
    )

    for RMSD, referenceRMSD in zip(RMSDs, referenceRMSDs):
        assert RMSD == pytest.approx(referenceRMSD)


@pytest.mark.parametrize(
    # Reference results obtained with OpenBabel
    "minimize, referenceRMSDs",
    [
        (
            True,  # Minimize: QCP + Isomorphism
            [
                0.476858,
                1.68089,
                1.50267,
                1.90623,
                1.01324,
                1.31716,
                1.11312,
                1.06044,
                0.965387,
                1.37842,
            ],
        ),
        (
            False,  # No minimize: Isomorphism only
            [
                0.592256,
                2.11545,
                2.29824,
                9.45773,
                1.35005,
                9.44356,
                9.59758,
                9.55076,
                2.44067,
                9.6171,
            ],
        ),
    ],
)
def test_rmsdwrapper_isomorphic(minimize: bool, referenceRMSDs: List[float]) -> None:

    molref = copy.deepcopy(molecules.docking_1cbr[0])
    mols = [copy.deepcopy(mol) for mol in molecules.docking_1cbr[1:]]

    RMSDs = spyrmsd.rmsdwrapper(molref, mols, minimize=minimize, strip=True)

    for RMSD, referenceRMSD in zip(RMSDs, referenceRMSDs):
        assert RMSD == pytest.approx(referenceRMSD, abs=1e-5)



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# Incase the project was not installed
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))


# -- Project information -----------------------------------------------------

project = "spyrmsd"
copyright = (
    "2019-2020, Rocco Meli. Project structure based on the "
    "Computational Molecular Science Python Cookiecutter version 1.1"
)
author = "Rocco Meli"

# The short X.Y version
# version = "0.0"
# The full version, including alpha/beta/rc tags
# release = "0.0.1"


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = "2.0"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autosummary",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
]

autosummary_generate = True
napoleon_google_docstring = False
napoleon_use_param = False
napoleon_use_ivar = True

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "Python"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "default"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"display_version": True}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "spyrmsddoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "spyrmsd.tex", "spyrmsd Documentation", "spyrmsd", "manual")
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "spyrmsd", "spyrmsd Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "spyrmsd",
        "spyrmsd Documentation",
        author,
        "spyrmsd",
        "Python tRMSD tool",
        "Miscellaneous",
    )
]


# -- Extension configuration -------------------------------------------------



import copy
import itertools
from typing import Tuple

import numpy as np
import pytest

from spyrmsd import molecule, qcp
from tests import molecules


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_M_mtx(mol: molecule.Molecule) -> None:
    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    # Build rotated coordinate set
    mol2.rotate(10, np.random.rand(3))

    M = qcp.M_mtx(mol1.coordinates, mol2.coordinates)

    assert M.shape == (3, 3)

    def S(i, j):
        A, B = mol1.coordinates, mol2.coordinates
        return np.dot(B[:, i], A[:, j])

    for i, j in itertools.combinations(range(3), 2):
        assert M[i, j] == pytest.approx(S(i, j))


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_K_mtx(mol: molecule.Molecule) -> None:

    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    # Build rotated coordinate set
    mol2.rotate(10, np.random.rand(3))

    M = qcp.M_mtx(mol1.coordinates, mol2.coordinates)
    K = qcp.K_mtx(M)

    assert K.shape == (4, 4)
    assert np.allclose(K.T, K)
    assert np.trace(K) == pytest.approx(0)


@pytest.mark.parametrize(
    "input, result",
    [
        ((2, 2, 0, 0, -1), 1),  # f(x) = x^4 - 1; x_0 = 2
        ((-2, -2, 0, 0, -1), -1),  # f(x) = x^4 - 1; x_0 = -2
        ((3, 3, -5, 0, 4), 2),  # f(x) = x^4 - 5 * x^2 + 4; x_0 = 3
        ((1, 1, -5, 0, 4), 1),  # f(x) = x^4 - 5 * x^2 + 4; x_0 = 1/2
        ((-1, -1, -5, 0, 4), -1),  # f(x) = x^4 - 5 * x^2 + 4; x_0 = -1/2
        ((-3, -3, -5, 0, 4), -2),  # f(x) = x^4 - 5 * x^2 + 4; x_0 = -3
    ],
)
def test_lambda_max(
    input: Tuple[float, float, float, float, float], result: float
) -> None:

    assert qcp.lambda_max(*input) == pytest.approx(result)



"""
Python RMSD tool with symmetry correction.
"""

from .due import due, Doi
from ._version import get_versions

versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions

# This will print latest Zenodo version
due.cite(
    Doi("10.5281/zenodo.3631876"),  # lgtm[py/procedure-return-value-used]
    path="spyrmsd",
    description="spyRMSD: Symmetry-Corrected RMSD in Python",
)



from typing import Any, List, Optional, Tuple, Union

import numpy as np

from spyrmsd import graph, hungarian, qcp, utils


def rmsd(
    coords1: np.ndarray,
    coords2: np.ndarray,
    atomicn1: np.ndarray,
    atomicn2: np.ndarray,
    center: bool = False,
    minimize: bool = False,
) -> float:
    """
    Compute RMSD

    Parameters
    ----------
    coords1: np.ndarray
        Coordinate of molecule 1
    coords2: np.ndarray
        Coordinates of molecule 2
    atomicn1: np.ndarray
        Atomic numbers for molecule 1
    atomicn2: np.ndarray
        Atomic numbers for molecule 2
    center: bool
        Center molecules at origin
    minimize: bool
        Compute minimum RMSD (with QCP method)

    Returns
    -------
    float
        RMSD

    Notes
    -----
    When `minimize=True`, the QCP method is used. [1]_ The molecules are
    centred at the origin according to the center of geometry and superimposed
    in order to minimize the RMSD.

    .. [1] D. L. Theobald, *Rapid calculation of RMSDs using a quaternion-based
       characteristic polynomial*, Acta Crys. A **61**, 478-480 (2005).
    """

    assert np.all(atomicn1 == atomicn2)
    assert coords1.shape == coords2.shape

    # Center coordinates if required
    c1 = utils.center(coords1) if center or minimize else coords1
    c2 = utils.center(coords2) if center or minimize else coords2

    if minimize:
        rmsd = qcp.qcp_rmsd(c1, c2)
    else:
        n = coords1.shape[0]

        rmsd = np.sqrt(np.sum((c1 - c2) ** 2) / n)

    return rmsd


def hrmsd(
    coords1: np.ndarray,
    coords2: np.ndarray,
    atomicn1: np.ndarray,
    atomicn2: np.ndarray,
    center=False,
):
    """
    Compute minimum RMSD using the Hungarian method.

    Parameters
    ----------
    coords1: np.ndarray
        Coordinate of molecule 1
    coords2: np.ndarray
        Coordinates of molecule 2
    atomicn1: np.ndarray
        Atomic numbers for molecule 1
    atomicn2: np.ndarray
        Atomic numbers for molecule 2

    Returns
    -------
    float
        Minimum RMSD (after assignment)

    Notes
    -----
    The Hungarian algorithm is used to solve the linear assignment problem, which is
    a minimum weight matching of the molecular graphs (bipartite). [2]_

    The linear assignment problem is solved for every element separately.

    .. [2] W. J. Allen and R. C. Rizzo, *Implementation of the Hungarian Algorithm to
        Account for Ligand Symmetry and Similarity in Structure-Based Design*,
        J. Chem. Inf. Model. **54**, 518-529 (2014)
    """

    assert atomicn1.shape == atomicn2.shape
    assert coords1.shape == coords2.shape

    # Center coordinates if required
    c1 = utils.center(coords1) if center else coords1
    c2 = utils.center(coords2) if center else coords2

    return hungarian.hungarian_rmsd(c1, c2, atomicn1, atomicn2)


def _rmsd_isomorphic_core(
    coords1: np.ndarray,
    coords2: np.ndarray,
    atomicnums1: np.ndarray,
    atomicnums2: np.ndarray,
    am1: np.ndarray,
    am2: np.ndarray,
    center: bool = False,
    minimize: bool = False,
    isomorphisms: Optional[List[Tuple[List[int], List[int]]]] = None,
) -> Tuple[float, List[Tuple[List[int], List[int]]]]:
    """
    Compute RMSD using graph isomorphism.

    Parameters
    ----------
    coords1: np.ndarray
        Coordinate of molecule 1
    coords2: np.ndarray
        Coordinates of molecule 2
    atomicnums1: npndarray
        Atomic numbers for molecule 1
    atomicnums2: npndarray
        Atomic numbers for molecule 2
    am1: np.ndarray
        Adjacency matrix for molecule 1
    am2: np.ndarray
        Adjacency matrix for molecule 2
    center: bool
        Centering flag
    minimize: bool
        Compute minized RMSD
    isomorphisms: Optional[List[Dict[int,int]]]
        Previously computed graph isomorphism

    Returns
    -------
    Tuple[float, List[Dict[int, int]]]
        RMSD (after graph matching) and graph isomorphisms
    """

    assert coords1.shape == coords2.shape

    n = coords1.shape[0]

    # Center coordinates if required
    c1 = utils.center(coords1) if center or minimize else coords1
    c2 = utils.center(coords2) if center or minimize else coords2

    # No cached isomorphisms
    if isomorphisms is None:
        # Convert molecules to graphs
        G1 = graph.graph_from_adjacency_matrix(am1, atomicnums1)
        G2 = graph.graph_from_adjacency_matrix(am2, atomicnums2)

        # Get all the possible graph isomorphisms
        isomorphisms = graph.match_graphs(G1, G2)

    # Minimum result
    # Squared displacement (not minimize) or RMSD (minimize)
    min_result = np.inf

    # Loop over all graph isomorphisms to find the lowest RMSD
    for idx1, idx2 in isomorphisms:

        # Use the isomorphism to shuffle coordinates around (from original order)
        c1i = c1[idx1, :]
        c2i = c2[idx2, :]

        if not minimize:
            # Compute square displacement
            # Avoid dividing by n and an expensive sqrt() operation
            result = np.sum((c1i - c2i) ** 2)
        else:
            # Compute minimized RMSD using QCP
            result = qcp.qcp_rmsd(c1i, c2i)

        min_result = result if result < min_result else min_result

    if not minimize:
        # Compute actual RMSD from square displacement
        min_result = np.sqrt(min_result / n)

    # Return the actual RMSD
    return min_result, isomorphisms


def symmrmsd(
    coordsref: np.ndarray,
    coords: Union[np.ndarray, List[np.ndarray]],
    atomicnumsref: np.ndarray,
    atomicnums: np.ndarray,
    amref: np.ndarray,
    am: np.ndarray,
    center: bool = False,
    minimize: bool = False,
    cache: bool = True,
) -> Any:
    """
    Compute RMSD using graph isomorphism for multiple coordinates.

    Parameters
    ----------
    coordsref: np.ndarray
        Coordinate of reference molecule
    coords: List[np.ndarray]
        Coordinates of other molecule
    atomicnumsref: npndarray
        Atomic numbers for reference
    atomicnums: npndarray
        Atomic numbers for other molecule
    amref: np.ndarray
        Adjacency matrix for reference molecule
    am: np.ndarray
        Adjacency matrix for other molecule
    center: bool
        Centering flag
    minimize: bool
        Minimum RMSD

    Returns
    -------
    float: Union[float, List[float]]
        Symmetry-corrected RMSD(s) and graph isomorphisms

    Notes
    -----

    Graph isomorphism is introduced for symmetry corrections. However, it is also
    useful when two molecules do not have the atoms in the same order since atom
    matching according to atomic numbers and the molecular connectivity is
    performed. If atoms are in the same order and there is no symmetry, use the
    `rmsd` function.
    """

    if isinstance(coords, list):  # Multiple RMSD calculations

        RMSD: Any = []
        isomorphism = None

        for c in coords:

            if not cache:
                # Reset isomorphism
                isomorphism = None

            srmsd, isomorphism = _rmsd_isomorphic_core(
                coordsref,
                c,
                atomicnumsref,
                atomicnums,
                amref,
                am,
                center=center,
                minimize=minimize,
                isomorphisms=isomorphism,
            )

            RMSD.append(srmsd)

    else:  # Single RMSD calculation

        RMSD, isomorphism = _rmsd_isomorphic_core(
            coordsref,
            coords,
            atomicnumsref,
            atomicnums,
            amref,
            am,
            center=center,
            minimize=minimize,
            isomorphisms=None,
        )

    return RMSD



try:
    from spyrmsd.optional.obabel import (
        load,
        loadall,
        adjacency_matrix,
        to_molecule,
        numatoms,
        numbonds,
        bonds,
    )

except ImportError:

    try:
        from spyrmsd.optional.rdkit import (
            load,
            loadall,
            adjacency_matrix,
            to_molecule,
            numatoms,
            numbonds,
            bonds,
        )
    except ImportError:
        # Use sPyRMSD as standalone library
        __all__ = []
    else:
        # Avoid flake8 complaint "imported but unused"
        __all__ = [
            "load",
            "loadall",
            "adjacency_matrix",
            "to_molecule",
            "numatoms",
            "numbonds",
            "bonds",
        ]
else:
    # Avoid flake8 complaint "imported but unused"
    __all__ = [
        "load",
        "loadall",
        "adjacency_matrix",
        "to_molecule",
        "numatoms",
        "numbonds",
        "bonds",
    ]

from typing import List

from spyrmsd import molecule


def loadmol(fname: str, adjacency: bool = True) -> molecule.Molecule:
    """
    Load molecule from file.

    Parameters
    ----------
    fname: str
        File name

    Returns
    -------
    molecule.Molecule
        `spyrmsd` molecule
    """

    mol = load(fname)

    return to_molecule(mol, adjacency=adjacency)


def loadallmols(fname: str, adjacency: bool = True) -> List[molecule.Molecule]:
    """
    Load molecules from file.

    Parameters
    ----------
    fname: str
        File name

    Returns
    -------
    List[molecule.Molecule]
        `spyrmsd` molecules
    """

    mols = loadall(fname)

    return [to_molecule(mol, adjacency=adjacency) for mol in mols]



import numpy as np
import scipy

from .due import due, Doi

due.cite(
    Doi("10.1021/ci400534h"),  # lgtm[py/procedure-return-value-used]
    path="spyrmsd.hungarian",
    description="Hungarian method",
)


def cost_mtx(A: np.ndarray, B: np.ndarray):
    """
    Compute the cost matrix for atom-atom assignment.

    Parameters
    ----------
    A: numpy.ndarray
        Atomic coordinates of molecule A
    B: numpy.ndarray
        Atomic coordinates of molecule B

    Returns
    -------
    np.ndarray
        Cost matrix of squared atomic distances between atoms of
        molecules A and B
    """

    return scipy.spatial.distance.cdist(A, B) ** 2


def optimal_assignment(A: np.ndarray, B: np.ndarray):
    """
    Solve the optimal assignment problems between atomic coordinates of
    molecules A and B.

    Parameters
    ----------
    A: numpy.ndarray
        Atomic coordinates of molecule A
    B: numpy.ndarray
        Atomic coordinates of molecule B

    Returns
    -------
    Tuple[float, nd.array, nd.array]
        Cost of the optimal assignment, together with the row and column
        indices of said assignment
    """

    C = cost_mtx(A, B)

    row_idx, col_idx = scipy.optimize.linear_sum_assignment(C)

    # Compute assignment cost
    cost = C[row_idx, col_idx].sum()

    return cost, row_idx, col_idx


def hungarian_rmsd(
    A: np.ndarray, B: np.ndarray, atomicnumsA: np.ndarray, atomicnumsB: np.ndarray
) -> float:
    """
    Solve the optimal assignment problems between atomic coordinates of
    molecules A and B.

    Parameters
    ----------
    A: numpy.ndarray
        Atomic coordinates of molecule A
    B: numpy.ndarray
        Atomic coordinates of molecule B
    atomicnumsA: numpy.ndarray
        Atomic numbers of molecule A
    atomicnumsB: numpy.ndarray
        Atomic numbers of molecule B

    Returns
    -------
    float
        RMSD computed with the Hungarian method

    Notes
    -----
    The Hungarian algorithm is used to solve the linear assignment problem, which is
    a minimum weight matching of the molecular graphs (bipartite).

    The linear assignment problem is solved for every element separately. [1]_

    .. [1] W. J. Allen and R. C. Rizzo, *Implementation of the Hungarian Algorithm to
        Account for Ligand Symmetry and Similarity in Structure-Based Design*,
        J. Chem. Inf. Model. **54**, 518-529 (2014)
    """

    assert A.shape == B.shape
    assert atomicnumsA.shape == atomicnumsB.shape

    elements = set(atomicnumsA)

    total_cost: float = 0.0
    for t in elements:
        atomicnumsA_idx = atomicnumsA == t
        atomicnumsB_idx = atomicnumsB == t

        assert atomicnumsA_idx.shape == atomicnumsB_idx.shape

        cost, row_idx, col_idx = optimal_assignment(
            A[atomicnumsA_idx, :], B[atomicnumsB_idx, :]
        )

        total_cost += cost

    N = A.shape[0]

    rmsd = np.sqrt(total_cost / N)

    return rmsd



import os
import re

import numpy as np
import pytest

from spyrmsd import io, rmsd

fdir = os.path.dirname(os.path.abspath(__file__))
molpath = os.path.join(fdir, "data", "docking")


def systems():
    dirs = os.listdir(molpath)

    # Match 4-character strings starting with a number
    systems = [d for d in dirs if re.match("[1-9]...", d)]

    return systems


@pytest.mark.slow
@pytest.fixture(scope="module", params=systems())
def molecules(request):
    system = request.param

    fref = os.path.join(molpath, system, f"{system}_ligand.sdf")
    fmols = os.path.join(molpath, system, f"{system}_dock.sdf")

    ref = io.load(os.path.join(molpath, fref))
    mols = io.loadall(os.path.join(molpath, fmols))

    ref = io.to_molecule(ref)
    mols = [io.to_molecule(mol) for mol in mols]

    ref.strip()
    for mol in mols:
        mol.strip()

    return ref, mols, system


@pytest.mark.slow
@pytest.mark.parametrize("cache", [True, False])
def test_benchmark_symmrmsd(cache, molecules, benchmark):

    ref, mols, system = molecules

    coords = [mol.coordinates for mol in mols]

    RMSDs = benchmark(
        rmsd.symmrmsd,
        ref.coordinates,
        coords,
        ref.atomicnums,
        mols[0].atomicnums,
        ref.adjacency_matrix,
        mols[0].adjacency_matrix,
        cache=cache,
    )

    testRMSDs = np.loadtxt(os.path.join(molpath, system, "obrms.dat"))

    assert np.allclose(RMSDs, testRMSDs)



from typing import List, Tuple

import numpy as np
import rdkit.Chem as Chem

from spyrmsd import molecule, utils


def load(fname: str):
    """
    Load molecule from file.

    Parameters
    ----------
    fname: str
        File name

    Returns
    -------
    Molecule
    """

    fmt = utils.molformat(fname)

    if fmt == "mol2":
        rdmol = Chem.MolFromMol2File(fname, removeHs=False)
    elif fmt == "sdf":
        rdmol = Chem.SDMolSupplier(fname, removeHs=False)[0]
    elif fmt == "pdb":
        rdmol = Chem.MolFromPDBFile(fname, removeHs=False)
    else:
        raise NotImplementedError

    return rdmol


def loadall(fname: str):
    """
    Load molecules from file.

    Parameters
    ----------
    fname: str
        File name

    Returns
    -------
    List of molecules
    """

    fmt = utils.molformat(fname)

    if fmt == "mol2":
        raise NotImplementedError  # See RDKit Issue #415
    elif fmt == "sdf":
        rdmols = Chem.SDMolSupplier(fname, removeHs=False)
    elif fmt == "pdb":
        # TODO: Implement
        raise NotImplementedError
    else:
        raise NotImplementedError

    return [rdmol for rdmol in rdmols]


def adjacency_matrix(mol) -> np.ndarray:
    """
    Adjacency matrix from OpenBabel molecule.

    Parameters
    ----------
    mol:
        Molecule

    Returns
    -------
    np.ndarray
        Adjacency matrix of the molecule
    """

    return Chem.rdmolops.GetAdjacencyMatrix(mol)


def to_molecule(mol, adjacency: bool = True) -> molecule.Molecule:
    """
    Transform molecule to `pyrmsd` molecule.

    Parameters
    ----------
    mol:
        Molecule
    adjacency: boolean, optional
        Flag to decide wether to build the adjacency matrix from molecule

    Returns
    -------
    spyrmsd.molecule.Molecule
        `spyrmsd` molecule
    """

    atoms = mol.GetAtoms()

    n = len(atoms)

    atomicnums = np.zeros(n, dtype=int)
    coordinates = np.zeros((n, 3))

    conformer = mol.GetConformer()

    for i, atom in enumerate(atoms):
        atomicnums[i] = atom.GetAtomicNum()

        pos = conformer.GetAtomPosition(i)

        coordinates[i] = np.array([pos.x, pos.y, pos.z])

    if adjacency:
        A = adjacency_matrix(mol)

    return molecule.Molecule(atomicnums, coordinates, A)


def numatoms(mol) -> int:
    """
    Number of atoms.

    Parameters
    ----------
    mol:
        Molecule

    Returns
    -------
    int
        Number of atoms
    """
    return mol.GetNumAtoms()


def numbonds(mol) -> int:
    """
    Number of bonds.

    Parameters
    ----------
    mol:
        Molecule

    Returns
    -------
    int
        Number of bonds
    """
    return mol.GetNumBonds()


def bonds(mol) -> List[Tuple[int, int]]:
    """
    List of bonds.

    Parameters
    ----------
    mol:
        Molecule

    Returns
    -------
    List[Tuple[int, int]]
        List of bonds

    Notes
    -----
    A bond is defined by a tuple of (0-based) indices of two atoms.
    """
    b = []

    for bond in mol.GetBonds():
        b.append((bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()))

    return b



"""
spyrmsd

Python tool for symmetry-corrected RMSD
"""

from setuptools import find_packages, setup

import versioneer

short_description = __doc__.split("\n")

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except Exception:
    long_description = "\n".join(short_description[2:])


setup(
    # Self-descriptive entries which should always be present
    name="spyrmsd",
    author="Rocco Meli",
    author_email="rocco.meli@biodtp.ox.ac.uk",
    description=short_description[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    url="https://spyrmsd.readthedocs.io",
    install_requires=["numpy", "scipy", "qcelemental", "networkx>=2"],
    extras_require={"bib": ["duecredit"]},
    platforms=["Linux", "Mac OS-X", "Unix", "Windows"],
    python_requires=">=3.6",
)



import copy
from typing import List

import numpy as np
import pytest

from spyrmsd import molecule, rmsd
from tests import molecules


@pytest.mark.parametrize("t, RMSD", [(0.0, 0.0), (1.0, 1.0), (2.0, 2.0)])
def test_rmsd_benzene(t: float, RMSD: float) -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.translate(np.array([0, 0, t]))

    assert rmsd.rmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(RMSD)


# Results obtained with PyTraj
#   pytraj.analysis.rmsd.rmsd(i, ref=j, nofit=True)
@pytest.mark.parametrize(
    "i, j, result", [(1, 2, 2.60065218), (1, 3, 9.94411523), (2, 3, 9.4091711)]
)
def test_rmsd_2viz(i: int, j: int, result: float) -> None:

    moli = copy.deepcopy(molecules.docking_2viz[i])
    molj = copy.deepcopy(molecules.docking_2viz[j])

    assert rmsd.rmsd(
        moli.coordinates, molj.coordinates, moli.atomicnums, molj.atomicnums
    ) == pytest.approx(result)


# Results obtained with PyTraj
#   pytraj.analysis.rmsd.rmsd(i, mask="!@H=", ref=j, ref_mask="!@H=", nofit=True)
@pytest.mark.parametrize(
    "i, j, result", [(1, 2, 2.65327362), (1, 3, 10.11099065), (2, 3, 9.57099612)]
)
def test_rmsd_2viz_stripped(i: int, j: int, result: float) -> None:

    moli = copy.deepcopy(molecules.docking_2viz[i])
    molj = copy.deepcopy(molecules.docking_2viz[j])

    moli.strip()
    molj.strip()

    assert rmsd.rmsd(
        moli.coordinates, molj.coordinates, moli.atomicnums, molj.atomicnums
    ) == pytest.approx(result)


def test_rmsd_centred_benzene() -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.translate(np.array([0, 0, 1]))

    assert rmsd.rmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(1)

    assert rmsd.rmsd(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        center=True,
    ) == pytest.approx(0)


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_rmsd_minimize(mol: molecule.Molecule) -> None:

    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    assert rmsd.rmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)

    assert rmsd.rmsd(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        minimize=True,
    ) == pytest.approx(0)

    for _ in range(10):
        mol2.translate(np.random.rand(3))
        mol2.rotate(np.random.rand(1), np.random.rand(3))

        assert (
            rmsd.rmsd(
                mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
            )
            > 0.0
        )

        assert rmsd.rmsd(
            mol1.coordinates,
            mol2.coordinates,
            mol1.atomicnums,
            mol2.atomicnums,
            minimize=True,
        ) == pytest.approx(0)


# Results obtained with PyTraj
#   pytraj.analysis.rmsd.rmsd(i, ref=j)
@pytest.mark.parametrize(
    "i, j, result", [(1, 2, 1.95277757), (1, 3, 3.11801105), (2, 3, 2.98609758)]
)
def test_rmsd_qcp_2viz(i: int, j: int, result: float) -> None:

    moli = copy.deepcopy(molecules.docking_2viz[i])
    molj = copy.deepcopy(molecules.docking_2viz[j])

    assert rmsd.rmsd(
        moli.coordinates,
        molj.coordinates,
        moli.atomicnums,
        molj.atomicnums,
        minimize=True,
    ) == pytest.approx(result)


# Results obtained with PyTraj
#   pytraj.analysis.rmsd.rmsd(i, "!@H=", ref=j, ref_mask="!@H=")
@pytest.mark.parametrize(
    "i, j, result", [(1, 2, 1.98171656), (1, 3, 3.01799306), (2, 3, 2.82917355)]
)
def test_rmsd_qcp_2viz_stripped(i: int, j: int, result: float) -> None:

    moli = copy.deepcopy(molecules.docking_2viz[i])
    molj = copy.deepcopy(molecules.docking_2viz[j])

    # Strip hydrogen atoms
    moli.strip()
    molj.strip()

    assert rmsd.rmsd(
        moli.coordinates,
        molj.coordinates,
        moli.atomicnums,
        molj.atomicnums,
        minimize=True,
    ) == pytest.approx(result)


# Results obtained with MDAnalysis
#   trp0 = mda.Universe("trp0.pdb")
#   c0 = trp0.coord -trp0.select_atoms("protein").center_of_geometry()
#   for i in [1,2,3,4,5]:
#       trp = mda.Universe(f"trp{i}.pdb")
#       rmsd_dummy = mda.analysis.rms.rmsd(trp.coord, trp0.coord)
#       c = trp.coord - trp.select_atoms("protein").center_of_geometry()
#       _, rmsd_min = align.rotation_matrix(tc, tc0)
#       print(rmsd_dummy, rmsd_min)
@pytest.mark.parametrize(
    "i, rmsd_dummy, rmsd_min",
    [
        (1, 4.812480551076202, 1.6578281551053196),
        (2, 6.772045449820714, 1.7175638492348284),
        (3, 9.344911262612964, 1.5946081072641485),
        (4, 9.772939589989000, 2.1234944939308220),
        (5, 8.901837608843241, 2.4894805175766606),
    ],
)
def test_rmsd_qcp_protein(i: int, rmsd_dummy: float, rmsd_min: float):

    mol0 = copy.deepcopy(molecules.trp[0])
    mol = copy.deepcopy(molecules.trp[i])

    assert rmsd.rmsd(
        mol0.coordinates, mol.coordinates, mol0.atomicnums, mol.atomicnums
    ) == pytest.approx(rmsd_dummy)

    assert rmsd.rmsd(
        mol0.coordinates,
        mol.coordinates,
        mol0.atomicnums,
        mol.atomicnums,
        minimize=True,
    ) == pytest.approx(rmsd_min)


@pytest.mark.parametrize(
    "angle, tol", [(60, 1e-4), (120, 1e-4), (180, 1e-4), (240, 1e-4), (300, 1e-4)]
)
def test_rmsd_hungarian_benzene_rotated(angle: float, tol: float) -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    assert rmsd.rmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)

    assert rmsd.hrmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)

    # Rotations different than 180 degrees introduce numerical errors (~1e-6)
    mol2.rotate(angle, [0, 0, 1], units="deg")

    assert (
        rmsd.rmsd(mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums)
        > 0
    )
    assert rmsd.hrmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0, abs=tol)


@pytest.mark.parametrize(
    "angle, tol", [(60, 1e-4), (120, 1e-4), (180, 1e-4), (240, 1e-4), (300, 1e-4)]
)
def test_rmsd_hungarian_benzene_shifted_rotated(angle: float, tol: float) -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.translate([0, 0, 1])

    assert rmsd.rmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(1)

    assert rmsd.hrmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(1)

    # Rotations different than 180 degrees introduce numerical errors (~1e-11)
    mol2.rotate(angle, [0, 0, 1], units="deg")

    assert (
        rmsd.rmsd(mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums)
        > 1
    )
    assert rmsd.hrmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(1, abs=tol)


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_rmsd_hungarian_centred(mol: molecule.Molecule) -> None:

    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    mol2.translate(np.random.rand(3))

    assert (
        rmsd.hrmsd(mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums)
        > 0
    )

    assert rmsd.hrmsd(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        center=True,
    ) == pytest.approx(0)


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_symmrmsd_centred(mol: molecule.Molecule) -> None:

    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    mol2.translate(np.random.rand(3))

    assert (
        rmsd.symmrmsd(
            mol1.coordinates,
            mol2.coordinates,
            mol1.atomicnums,
            mol2.atomicnums,
            mol1.adjacency_matrix,
            mol2.adjacency_matrix,
        )
        > 0
    )

    assert rmsd.symmrmsd(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        mol1.adjacency_matrix,
        mol2.adjacency_matrix,
        center=True,
    ) == pytest.approx(0)


@pytest.mark.parametrize("angle", [60, 120, 180, 240, 300, 360])
def test_symmrmsd_rotated_benzene(angle: float) -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.rotate(angle, np.array([0, 0, 1]), units="deg")

    assert (
        rmsd.rmsd(mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums)
        > 0
    )

    assert rmsd.rmsd(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        minimize=True,
    ) == pytest.approx(0)

    assert rmsd.hrmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0, abs=1e-4)

    assert rmsd.symmrmsd(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        mol1.adjacency_matrix,
        mol2.adjacency_matrix,
    ) == pytest.approx(0, abs=1e-4)


@pytest.mark.parametrize("angle", [60, 120, 180, 240, 300, 360])
def test_symmrmsd_rotated_benzene_stripped(angle: float) -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.rotate(angle, np.array([0, 0, 1]), units="deg")

    mol1.strip()
    mol2.strip()

    assert (
        rmsd.rmsd(mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums)
        > 0
    )

    assert rmsd.rmsd(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        minimize=True,
    ) == pytest.approx(0)

    assert rmsd.hrmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0, abs=1e-4)

    assert rmsd.symmrmsd(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        mol1.adjacency_matrix,
        mol2.adjacency_matrix,
    ) == pytest.approx(0, abs=1e-4)


def test_symmrmsd_atomicnums_matching_pyridine_stripped() -> None:

    mol1 = copy.deepcopy(molecules.pyridine)
    mol2 = copy.deepcopy(molecules.pyridine)

    mol2.rotate(60, np.array([0, 0, 1]), units="deg")

    mol1.strip()
    mol2.strip()

    # Standard RMSD, correct in this case
    RMSD = rmsd.rmsd(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    )

    # Isomorphic RMSD with atomic number matching is correct
    # Without atomic number matching this would be wrong because of higher symmetry
    assert rmsd.symmrmsd(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        mol1.adjacency_matrix,
        mol2.adjacency_matrix,
    ) == pytest.approx(RMSD, abs=1e-4)


# Results obtained with OpenBabel
@pytest.mark.parametrize(
    "index, RMSD, minimize",
    [
        (1, 0.592256, False),
        (2, 2.11545, False),
        (3, 2.29824, False),
        (4, 9.45773, False),
        (5, 1.35005, False),
        (6, 9.44356, False),
        (7, 9.59758, False),
        (8, 9.55076, False),
        (9, 2.44067, False),
        (10, 9.6171, False),
        (1, 0.476858, True),
        (2, 1.68089, True),
        (3, 1.50267, True),
        (4, 1.90623, True),
        (5, 1.01324, True),
        (6, 1.31716, True),
        (7, 1.11312, True),
        (8, 1.06044, True),
        (9, 0.965387, True),
        (10, 1.37842, True),
    ],
)
def test_rmsd_symmrmsd(index: int, RMSD: float, minimize: bool) -> None:

    molc = copy.deepcopy(molecules.docking_1cbr[0])
    mol = copy.deepcopy(molecules.docking_1cbr[index])

    molc.strip()
    mol.strip()

    assert rmsd.symmrmsd(
        molc.coordinates,
        mol.coordinates,
        molc.atomicnums,
        mol.atomicnums,
        molc.adjacency_matrix,
        mol.adjacency_matrix,
        minimize=minimize,
    ) == pytest.approx(RMSD, abs=1e-5)


# Results obtained with OpenBabel
@pytest.mark.parametrize(
    "minimize, referenceRMSDs",
    [
        (
            False,
            [
                0.592256,
                2.11545,
                2.29824,
                9.45773,
                1.35005,
                9.44356,
                9.59758,
                9.55076,
                2.44067,
                9.6171,
            ],
        ),
        (
            True,
            [
                0.476858,
                1.68089,
                1.50267,
                1.90623,
                1.01324,
                1.31716,
                1.11312,
                1.06044,
                0.965387,
                1.37842,
            ],
        ),
    ],
)
def test_multirmsd_isomorphic(minimize: bool, referenceRMSDs: List[float]) -> None:

    molc = copy.deepcopy(molecules.docking_1cbr[0])
    mols = [copy.deepcopy(mol) for mol in molecules.docking_1cbr[1:]]

    molc.strip()

    for mol in mols:
        mol.strip()

    RMSDs = rmsd.symmrmsd(
        molc.coordinates,
        [mol.coordinates for mol in mols],
        molc.atomicnums,
        mols[0].atomicnums,
        molc.adjacency_matrix,
        mols[0].adjacency_matrix,
        minimize=minimize,
    )

    for RMSD, referenceRMSD in zip(RMSDs, referenceRMSDs):
        assert RMSD == pytest.approx(referenceRMSD, abs=1e-5)


# Results obtained with OpenBabel
@pytest.mark.parametrize(
    "minimize, referenceRMSDs",
    [
        (
            False,
            [
                0.592256,
                2.11545,
                2.29824,
                9.45773,
                1.35005,
                9.44356,
                9.59758,
                9.55076,
                2.44067,
                9.6171,
            ],
        ),
        (
            True,
            [
                0.476858,
                1.68089,
                1.50267,
                1.90623,
                1.01324,
                1.31716,
                1.11312,
                1.06044,
                0.965387,
                1.37842,
            ],
        ),
    ],
)
def test_symmrmsd_cache(minimize: bool, referenceRMSDs: List[float]) -> None:

    molc = copy.deepcopy(molecules.docking_1cbr[0])
    mols = [copy.deepcopy(mol) for mol in molecules.docking_1cbr[1:]]

    molc.strip()

    for mol in mols:
        mol.strip()

    RMSDs = rmsd.symmrmsd(
        molc.coordinates,
        [mol.coordinates for mol in mols],
        molc.atomicnums,
        mols[0].atomicnums,
        molc.adjacency_matrix,
        mols[0].adjacency_matrix,
        minimize=minimize,
        cache=False,
    )

    for RMSD, referenceRMSD in zip(RMSDs, referenceRMSDs):
        assert RMSD == pytest.approx(referenceRMSD, abs=1e-5)



import os
from typing import Any, List, Tuple

from spyrmsd import io, molecule

fdir = os.path.dirname(os.path.abspath(__file__))
molpath = os.path.join(fdir, "data/molecules/")


def load(fname: str) -> Tuple[Any, molecule.Molecule]:
    """
    Load molecule from file.

    Parameters
    ----------
    fname: str
        Input file name

    Returns
    -------
    Tuple[Any, molecule.Molecule]
        Loaded molecule as `pybel.Molecule` or `rdkit.Chem.rdkem.Mol` and
        `pyrmsd.molecule.Molecule`
    """

    fname = os.path.join(molpath, fname)

    m = io.load(fname)

    mol = io.to_molecule(m, adjacency=True)

    return m, mol


def loadall(fname: str) -> Tuple[List[Any], List[molecule.Molecule]]:
    """
    Load all molecule from file.

    Parameters
    ----------
    fname: str
        Input file name

    Returns
    -------
    Tuple[List[Any], List[molecule.Molecule]]
        Loaded molecule as `pybel.Molecule` or `rdkit.Chem.rdchem.Mol` and
        `pyrmsd.molecule.Molecule`
    """

    fname = os.path.join(molpath, fname)

    ms = io.loadall(fname)

    mols = [io.to_molecule(m, adjacency=True) for m in ms]

    return ms, mols


obbenzene, benzene = load("benzene.sdf")
obpyridine, pyridine = load("pyridine.sdf")
obethanol, ethanol = load("ethanol.sdf")
obdialanine, dialanine = load("dialanine.sdf")
obsdf = [obbenzene, obpyridine, obethanol, obdialanine]
sdf = [benzene, pyridine, ethanol, dialanine]

allmolecules = sdf
allobmolecules = obsdf

obdocking_2viz, docking_2viz = {}, {}
for i in [1, 2, 3]:
    obdocking_2viz[i], docking_2viz[i] = load(f"2viz_{i}.sdf")

obdocking_1cbr = [load("1cbr_ligand.mol2")[0], *loadall("1cbr_docking.sdf")[0]]
docking_1cbr = [load("1cbr_ligand.mol2")[1], *loadall("1cbr_docking.sdf")[1]]

intrp, trp = [], []
for i in range(6):
    intrp.append(load(f"trp{i}.pdb")[0])
    trp.append(load(f"trp{i}.pdb")[1])



import os

import numpy as np


def format(fname: str) -> str:
    """
    Extract format extension from file name.

    Parameters
    ----------
    fname : str
        File name

    Returns
    -------
    str
        File extension

    Notes
    -----
    The file extension is returned without the `.` character, i.e. for the file
    `path/filename.ext` the string `ext` is returned.
    """
    name, ext = os.path.splitext(fname)

    return ext[1:]  # Remove "."


def molformat(fname: str) -> str:
    """
    Extract an OpenBabel-friendly format from file name.

    Parameters
    ----------
    fname : str
        File name

    Returns
    -------
    str
        File extension in an OpenBabel-friendly format

    Notes
    -----
    File types in OpenBabel do not always correspond to the file extension. This
    function converts the file extension to an OpenBabel file type.

    The following table shows the different conversions performed by this function:

    ========= =========
    Extension File Type
    --------- ---------
    xyz       XYZ
    ========= =========
    """

    ext = format(fname)

    if ext == "xyz":
        # xyz files in OpenBabel are called XYZ
        ext = ext.upper()

    return ext


def deg_to_rad(angle: float) -> float:
    """
    Convert angle in degrees to angle in radians.

    Parameters
    ----------
    angle : float
        Angle (in degrees)

    Returns
    -------
    float
        Angle (in radians)
    """

    return angle * np.pi / 180.0


def rotate(v: np.array, angle: float, axis: np.array, units: str = "rad") -> np.array:
    """
    Rotate vector.

    Parameters
    ----------
    v: numpy.array
        3D vector to be rotated
    angle : float
        Angle of rotation (in `units`)
    axis : numpy.array
        3D axis of rotation
    units: {"rad", "deg"}
        Units of `angle` (in radians `rad` or degrees `deg`)

    Returns
    -------
    numpy.array
        Rotated vector

    Raises
    ------
    AssertionError
        If the axis of rotation is not a 3D vector
    ValueError
        If `units` is not `rad` or `deg`
    """

    assert len(axis) == 3

    # Ensure rotation axis is normalised
    axis = axis / np.linalg.norm(axis)

    if units.lower() == "rad":
        pass
    elif units.lower() == "deg":
        angle = deg_to_rad(angle)
    else:
        raise ValueError(
            f"Units {units} for angle is not supported. Use 'deg' or 'rad' instead."
        )

    t1 = np.outer(axis, np.inner(axis, v)).T
    t2 = np.cos(angle) * np.cross(np.cross(axis, v), axis)
    t3 = np.sin(angle) * np.cross(axis, v)

    return t1 + t2 + t3


def center_of_geometry(coordinates: np.ndarray) -> np.ndarray:
    """
    Center of geometry.

    Parameters
    ----------
    coordinates: np.ndarray
        Coordinates

    Returns
    -------
    np.ndarray
        Center of geometry
    """

    assert coordinates.shape[1] == 3

    return np.mean(coordinates, axis=0)


def center(coordinates: np.ndarray) -> np.ndarray:
    """
    Center coordinates.

    Parameters
    ----------
    coordinates: np.ndarray
        Coordinates

    Returns
    -------
    np.ndarray
        Centred coordinates
    """

    return coordinates - center_of_geometry(coordinates)



import warnings
from typing import Any, List, Optional, Tuple, Union

import networkx as nx
import numpy as np


def graph_from_adjacency_matrix(
    adjacency_matrix: Union[np.ndarray, List[List[int]]],
    atomicnums: Optional[Union[np.ndarray, List[int]]] = None,
) -> nx.Graph:
    """
    Graph from adjacency matrix.

    Parameters
    ----------
    adjacency_matrix: Union[np.ndarray, List[List[int]]]
        Adjacency matrix
    atomicnums: Union[np.ndarray, List[int]], optional
        Atomic numbers

    Returns
    -------
    Graph
        Molecular graph

    Notes
    -----
    It the atomic numbers are passed, they are used as node attributes.
    """

    G = nx.Graph(adjacency_matrix)

    if atomicnums is not None:
        attributes = {idx: atomicnum for idx, atomicnum in enumerate(atomicnums)}
        nx.set_node_attributes(G, attributes, "atomicnum")

    return G


def match_graphs(G1, G2) -> List[Tuple[List[int], List[int]]]:
    """
    Compute graph isomorphisms.

    Parameters
    ----------
    G1:
        Graph 1
    G2:
        Graph 2

    Returns
    -------
    List[Tuple[List[int],List[int]]]
        All possible mappings between nodes of graph 1 and graph 2 (isomorphisms)

    Raises
    ------
    ValueError
        If the graphs `G1` and `G2` are not isomorphic
    """

    def match_atomicnum(node1, node2):
        return node1["atomicnum"] == node2["atomicnum"]

    if (
        nx.get_node_attributes(G1, "atomicnum") == {}
        or nx.get_node_attributes(G2, "atomicnum") == {}
    ):
        # Nodes without atomic number information
        # No node-matching check
        node_match = None

        warnings.warn(
            "No atomic number information stored on nodes. "
            + "Node matching is not performed..."
        )

    else:
        node_match = match_atomicnum

    GM = nx.algorithms.isomorphism.GraphMatcher(G1, G2, node_match)

    # Check if graphs are actually isomorphic
    if not GM.is_isomorphic():
        # TODO: Create a new exception
        raise ValueError(f"Graphs {G1} and {G2} are not isomorphic.")

    return [
        (list(isomorphism.keys()), list(isomorphism.values()))
        for isomorphism in GM.isomorphisms_iter()
    ]


def vertex_property(G, vproperty: str, idx: int) -> Any:
    """
    Get vertex (node) property from graph

    Parameters
    ----------
    G:
        Graph
    vproperty: str
        Vertex property name
    idx: int
        Vertex index

    Returns
    -------
    Any
        Vertex property value
    """
    return G.nodes[idx][vproperty]


def num_vertices(G) -> int:
    """
    Number of vertices

    Parameters
    ----------
    G:
        Graph

    Returns
    -------
    int
        Number of vertices (nodes)
    """
    return G.number_of_nodes()


def num_edges(G) -> int:
    """
    Number of edges

    Parameters
    ----------
    G:
        Graph

    Returns
    -------
    int
        Number of edges
    """
    return G.number_of_edges()


def lattice(n1, n2):
    """
    Build 2D lattice graph

    Parameters
    ----------
    n1: int
        Number of nodes in dimension 1
    n2: int
        Number of nodes in dimension 2

    Returns
    -------
    Graph
        Lattice graph
    """
    return nx.generators.lattice.grid_2d_graph(n1, n2)


def cycle(n):
    """
    Build cycle graph

    Parameters
    ----------
    n: int
        Number of nodes

    Returns
    -------
    Graph
        Cycle graph
    """
    return nx.cycle_graph(n)



"""
Test imported package and modules.

This test is useful to generate a citation report:

.. code

   python -m duecite test_import.py

"""

import sys

import pytest

from spyrmsd import graph, hungarian, io, molecule, qcp, rmsd, spyrmsd, utils


@pytest.mark.parametrize(
    "name",
    [
        # Package
        spyrmsd.__name__,
        # Modules
        graph.__name__,
        hungarian.__name__,
        io.__name__,
        molecule.__name__,
        qcp.__name__,
        rmsd.__name__,
        spyrmsd.__name__,
        utils.__name__,
    ],
)
def test_imported(name):
    assert name in sys.modules
